#ifndef MSP_GEOMETRY_TRANSFORMEDSHAPE_H_
#define MSP_GEOMETRY_TRANSFORMEDSHAPE_H_

#include "affinetransformation.h"
#include "shape.h"

namespace Msp {
namespace Geometry {

/**
A shape modified by an affine transformation.
*/
template<typename T, unsigned D>
class TransformedShape: public Shape<T, D>
{
private:
	Shape<T, D> *shape;
	AffineTransformation<T, D> transformation;
	AffineTransformation<T, D> inverse_trans;

public:
	TransformedShape(const Shape<T, D> &, const AffineTransformation<T, D> &);
	TransformedShape(const TransformedShape &);
	TransformedShape &operator=(const TransformedShape &);
	~TransformedShape();

	virtual TransformedShape *clone() const;

	const Shape<T, D> &get_shape() const { return *shape; }
	const AffineTransformation<T, D> &get_transformation() const { return transformation; }

	virtual BoundingBox<T, D> get_axis_aligned_bounding_box() const;
	virtual bool contains(const LinAl::Vector<T, D> &) const;
	virtual unsigned get_max_ray_intersections() const { return shape->get_max_ray_intersections(); }
	virtual unsigned get_intersections(const Ray<T, D> &, SurfacePoint<T, D> *, unsigned) const;
};

template<typename T, unsigned D>
inline TransformedShape<T, D>::TransformedShape(const Shape<T, D> &s, const AffineTransformation<T, D> &t):
	shape(s.clone()),
	transformation(t),
	inverse_trans(invert(t))
{ }

template<typename T, unsigned D>
inline TransformedShape<T, D>::TransformedShape(const TransformedShape<T, D> &other):
	shape(other.shape->clone()),
	transformation(other.transformation),
	inverse_trans(other.inverse_trans)
{ }

template<typename T, unsigned D>
inline TransformedShape<T, D> &TransformedShape<T, D>::operator=(const TransformedShape<T, D> &other)
{
	delete shape;
	shape = other.shape->clone();
	transformation = other.transformation;
	inverse_trans = other.inverse_trans;
}

template<typename T, unsigned D>
inline TransformedShape<T, D>::~TransformedShape()
{
	delete shape;
}

template<typename T, unsigned D>
inline TransformedShape<T, D> *TransformedShape<T, D>::clone() const
{
	return new TransformedShape<T, D>(*this);
}

template<typename T, unsigned D>
inline BoundingBox<T, D> TransformedShape<T, D>::get_axis_aligned_bounding_box() const
{
	return transformation.transform(shape->get_axis_aligned_bounding_box());
}

template<typename T, unsigned D>
inline bool TransformedShape<T, D>::contains(const LinAl::Vector<T, D> &point) const
{
	return shape->contains(inverse_trans.transform(point));
}

template<typename T, unsigned D>
inline unsigned TransformedShape<T, D>::get_intersections(const Ray<T, D> &ray, SurfacePoint<T, D> *points, unsigned size) const
{
	Ray<T, D> local_ray = inverse_trans.transform(ray);

	unsigned count = shape->get_intersections(local_ray, points, size);
	if(points)
	{
		for(unsigned i=0; i<count; ++i)
		{
			points[i].position = transformation.transform(points[i].position);
			/* XXX This is not correct for nonuniform scaling.  Inverse of the
			transpose of the upper DxD part of the matrix should be used. */
			points[i].normal = transformation.transform_linear(points[i].normal);
			points[i].distance = inner_product(points[i].position-ray.get_start(), ray.get_direction());
		}
	}
	return count;
}

} // namespace Geometry
} // namespace Msp

#endif
